#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Author: Riku Laine
Date: 12AUG2019 / Modified 30DEC2019
Project name: Potential outcomes in model evaluation
Description: This script creates the figures and results used
             in empirical data experiments with COMPAS data.

Parameters:
-----------
(1) save_name : file path and prefix for saving the created figures.
(2) stan_code_file_name : Name of file containing the stan model code.
(3) sigma_tau : Values of prior variance for the Jung-inspired model.
(4) data_path : Path to data file.
(5) nJudges_M : Number of judges in simulations.
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import numpy.random as npr
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
import gc
import psutil
import os
from cmdstanpy import Model

plt.switch_backend('agg')

plt.rcParams.update({'font.size': 16})
plt.rcParams.update({'figure.figsize': (10, 6)})

import sys

# Seed for reproducibility
npr.seed(111)

# figure storage name
save_name = sys.argv[1] + "sl_compas_"

# Name of stan model code file
stan_code_file_name = sys.argv[2]

# Variance prior
sigma_tau = float(sys.argv[3])

# figure storage name
data_path = sys.argv[4]

nJudges_M = int(sys.argv[5])

# Prefix for the figures and log files

print("These results have been obtained with the following settings:")

print("Prior for the variances:", sigma_tau)

def standardize(x):
    '''Standardization function'''

    return (x - np.mean(x)) / np.std(x, ddof=1)

##########################

# ## Evaluator modules

# ### Convenience functions

def fitPredictiveModel(x_train, y_train, x_test, class_value, model_type=None):
    '''
    Fit a predictive model (default logistic regression) with given training
    instances and return probabilities for test instances to obtain a given
    class label.

    Arguments:
    ----------

    x_train -- x values of training instances
    y_train -- y values of training instances
    x_test -- x values of test instances
    class_value -- class label for which the probabilities are counted for.
    model_type -- type of model to be fitted.

    Returns:
    --------
    (1) Trained predictive model
    (2) Probabilities for given test inputs for given class.
    '''

    if model_type is None or model_type in ["logistic_regression", "lr"]:
        # Instantiate the model (using the default parameters)
        logreg = LogisticRegression(solver='lbfgs')

        # Check shape and fit the model.
        if x_train.ndim == 1:
            logreg = logreg.fit(x_train.values.reshape(-1, 1), y_train)
        else:
            logreg = logreg.fit(x_train, y_train)

        label_probs_logreg = getProbabilityForClass(x_test, logreg,
                                                    class_value)

        return logreg, label_probs_logreg

    elif model_type in ["random_forest", "rf"]:
        # Instantiate the model
        forest = RandomForestClassifier(n_estimators=100, max_depth=3)

        # Check shape and fit the model.
        if x_train.ndim == 1:
            forest = forest.fit(x_train.values.reshape(-1, 1), y_train)
        else:
            forest = forest.fit(x_train, y_train)

        label_probs_forest = getProbabilityForClass(x_test, forest,
                                                    class_value)

        return forest, label_probs_forest

    elif model_type == "fully_random":

        label_probs = np.ones_like(x_test) / 2

        model_object = lambda x: 0.5

        return model_object, label_probs
    else:
        raise ValueError("Invalid model_type!", model_type)


def getProbabilityForClass(x, model, class_value):
    '''
    Function (wrapper) for obtaining the probability of a class given x and a
    predictive model.

    Arguments:
    -----------
    x -- individual features, an array of shape (observations, features)
    model -- a trained sklearn model. Predicts probabilities for given x.
        Should accept input of shape (observations, features)
    class_value -- the resulting class to predict (usually 0 or 1).

    Returns:
    --------
    (1) The probabilities of given class label for each x.
    '''
    if x.ndim == 1:
        # if x is vector, transform to column matrix.
        f_values = model.predict_proba(np.array(x).reshape(-1, 1))
    else:
        f_values = model.predict_proba(x)

    # Get correct column of predicted class, remove extra dimensions and return.
    return f_values[:, model.classes_ == class_value].flatten()


# ### Contraction algorithm
#
# Below is an implementation of Lakkaraju's team's algorithm presented in
# [their paper](https://helka.finna.fi/PrimoRecord/pci.acm3098066). Relevant
# parameters to be passed to the function are presented in the description.

def contraction(df, judgeIDJ_col, decisionT_col, resultY_col, modelProbS_col,
                accRateR_col, r):
    '''
    This is an implementation of the algorithm presented by Lakkaraju
    et al. in their paper "The Selective Labels Problem: Evaluating
    Algorithmic Predictions in the Presence of Unobservables" (2017).

    Arguments:
    ----------
    df -- The (Pandas) data frame containing the data, judge decisions,
        judge IDs, results and probability scores.
    judgeIDJ_col -- String, the name of the column containing the judges' IDs
        in df.
    decisionT_col -- String, the name of the column containing the judges' decisions
    resultY_col -- String, the name of the column containing the realization
    modelProbS_col -- String, the name of the column containing the probability
        scores from the black-box model B.
    accRateR_col -- String, the name of the column containing the judges'
        acceptance rates
    r -- Float between 0 and 1, the given acceptance rate.

    Returns:
    --------
    (1) The estimated failure rate at acceptance rate r.
    '''
    # Get ID of the most lenient judge.
    most_lenient_ID_q = df[judgeIDJ_col].loc[df[accRateR_col].idxmax()]

    # Subset. "D_q is the set of all observations judged by q."
    D_q = df[df[judgeIDJ_col] == most_lenient_ID_q].copy()

    ##### Agreement rate computation begins #######

    Q_leniency = max(df[accRateR_col])

    if Q_leniency < r:
        return np.nan, np.nan

    # Sort D_q
    # "Observations deemed as high risk by B are at the top of this list"
    D_sort_q = D_q.sort_values(by=modelProbS_col, ascending=False)

    # The agreement rate is now computed as the percentage of all the positive
    # decisions given by q in the 1-r % of the most dangerous subjects.
    all_negatives = np.sum(D_sort_q[decisionT_col] == 0)

    n_most_dangerous = int(round(D_q.shape[0] * (1-Q_leniency)))

    agreed_decisions = np.sum(D_sort_q[decisionT_col][0:n_most_dangerous] == 0)

    print("The agreement rate of contraction is:", agreed_decisions/all_negatives)

    ##### Agreement rate computation ends #######

    # All observations of R_q have observed outcome labels.
    # "R_q is the set of observations in D_q with observed outcome labels."
    R_q = D_q[D_q[decisionT_col] == 1].copy()

    # Sort observations in R_q in descending order of confidence scores S and
    # assign to R_sort_q.
    # "Observations deemed as high risk by B are at the top of this list"
    R_sort_q = R_q.sort_values(by=modelProbS_col, ascending=False)

    number_to_remove = int(
        round((1.0 - r) * D_q.shape[0] - (D_q.shape[0] - R_q.shape[0])))

    # "R_B is the list of observations assigned to t = 1 by B"
    R_B = R_sort_q[number_to_remove:R_sort_q.shape[0]]

    return np.sum(R_B[resultY_col] == 0) / D_q.shape[0], D_q.shape[0]


# ### Evaluators

def trueEvaluationEvaluator(df, resultY_col, r):

    df.sort_values(by='B_prob_0_model', inplace=True, ascending=True)

    to_release = int(round(df.shape[0] * r))

    failed = df[resultY_col][0:to_release] == 0

    return np.sum(failed) / df.shape[0], df.shape[0]


def labeledOutcomesEvaluator(df, decisionT_col, resultY_col, r, adjusted=False):

    df_observed = df.loc[df[decisionT_col] == 1, :]

    df_observed = df_observed.sort_values(by='B_prob_0_model',
                                              inplace=False,
                                              ascending=True)

    to_release = int(round(df_observed.shape[0] * r))

    failed = df_observed[resultY_col][0:to_release] == 0

    if adjusted:
        return np.mean(failed), df.shape[0]

    return np.sum(failed) / df.shape[0], df.shape[0]


################### Data preprocessing ######################

# Read in the data
compas_raw = pd.read_csv(data_path)

# Select columns
compas = compas_raw[['age', 'c_charge_degree', 'race', 'age_cat', 'score_text',
                     'sex','priors_count', 'days_b_screening_arrest',
                     'decile_score', 'is_recid', 'two_year_recid', 'c_jail_in',
                     'c_jail_out']]

# Filter values, see ProPublica methodology.
compas = compas.query('days_b_screening_arrest <= 30 and \
                      days_b_screening_arrest >= -30 and \
                      is_recid != -1 and \
                      c_charge_degree != "O"')

# Drop row if score_text is na
compas = compas[compas.score_text.notnull()]

# Recode recidivism values to fit earlier notation
# So here result_Y = 1 - is_recid (inverted binary coding).
compas['result_Y'] = np.where(compas['is_recid'] == 1, 0, 1)

# Convert string values to dummies, drop first to keep full rank.
compas_dummy = pd.get_dummies(compas,
                              columns=['c_charge_degree', 'race', 'age_cat', 'sex'],
                              drop_first=True)

compas_dummy.drop(columns=['age', 'days_b_screening_arrest', 'c_jail_in',
                           'c_jail_out', 'two_year_recid', 'score_text',
                           'is_recid'], inplace=True)

# Shuffle rows for random judge assignment
compas_shuffled = compas_dummy.sample(frac=1)

# Decision assignment

# Generate leniencies from U(0.1, 0.9) for all 16 judges. Then distribute the
# observations evenly to all decision-makers and give positive decisions to
# r% of subjects with the lowest COMPAS score.

# nJudges_M = 16
# nJudges_M = 12

# Assign judges as evenly as possible
judge_ID = pd.qcut(np.arange(len(compas_shuffled)), nJudges_M, labels=False)

# Draw leniencies
# judge_leniency = npr.uniform(0.1, 0.9, nJudges_M)

# Assign leniencies
judge_leniency = np.repeat([0.1, 0.5, 0.9], int(nJudges_M/3))

# Replicate values for assignment
judge_leniency = judge_leniency[judge_ID]

# Attach to data
compas_shuffled = compas_shuffled.assign(judge_ID=judge_ID,
                                         judge_leniency=judge_leniency)

# Sort by judges then probabilities in decreasing order
# Most dangerous for each judge are at the top.
compas_shuffled.sort_values(by=["judge_ID", "decile_score"],
                            ascending=False,
                            inplace=True)

# Iterate over the data. Subject will be given a negative decision
# if they are in the top (1-r)*100% of the individuals the judge will judge.
# I.e. if their within-judge-index is under 1 - acceptance threshold times
# the number of subjects assigned to each judge they will receive a
# negative decision.
compas_shuffled.reset_index(drop=True, inplace=True)

subjects_allocated = compas_shuffled.judge_ID.value_counts()

compas_shuffled['judge_index'] = compas_shuffled.groupby('judge_ID').cumcount()

compas_shuffled['decision_T'] = np.where(
    compas_shuffled['judge_index'] < (1 - compas_shuffled['judge_leniency']) *
    subjects_allocated[compas_shuffled['judge_ID']].values, 0, 1)

compas_labeled = compas_shuffled.copy()
compas_unlabeled = compas_shuffled.copy()

# Hide unobserved
compas_labeled.loc[compas_labeled.decision_T == 0, 'result_Y'] = np.nan

# Choose feature_columns for the regression (priors, charge degree, race, age
# and sex).
feature_cols = ~compas_labeled.columns.isin(
    ['result_Y', 'decile_score', 'judge_ID', 'judge_leniency', 'judge_index', 'decision_T'])

feature_cols = compas_labeled.columns[feature_cols]

# Standardize priors count

compas_labeled.priors_count = standardize(compas_labeled.priors_count)

####### Draw figures ###########

def drawDiagnostics(title, save_name, f_rates, titles):
    '''
    Draw boxplot figures. One panel for each method.
    '''

    cols = 2
    rows = np.ceil(len(f_rates) / cols)

    plt.figure(figsize=(16, 4.5*rows+1))

    ax = plt.subplot(rows, cols, 1)
    x_ax = np.arange(1, 9, 1) / 10

    plt.boxplot(f_rates[0], labels=x_ax)

    plt.title(titles[0])
    plt.xlabel('Acceptance rate')
    plt.ylabel('Failure rate')
    plt.grid()

    for i in range(len(f_rates)):
        plt.subplot(rows, cols, i + 1, sharey=ax)

        plt.boxplot(f_rates[i], labels=x_ax)

        plt.title(titles[i])
        plt.xlabel('Acceptance rate')
        plt.ylabel('Failure rate')
        plt.grid()

    plt.tight_layout()
    plt.subplots_adjust(top=0.85)
    plt.suptitle(title, y=0.96, weight='bold')

    plt.savefig(save_name + '_diagnostic_plot')

    plt.show()
    plt.close('all')


nIter = 10

failure_rates = np.zeros((8, 4))
failure_stds = np.zeros((8, 4))

f_rate_true = np.zeros((nIter, 8))
f_rate_label = np.zeros((nIter, 8))
f_rate_cont = np.zeros((nIter, 8))
f_rate_cf = np.zeros((nIter, 8))

# Compile stan model
sm = Model(stan_file=stan_code_file_name)
sm.compile()

# Split the nJudges_M judges to two datasets nIter times.

for i in range(nIter):

    # Split data using judge IDs.

    # Get number of judges
    id_max = np.max(compas_labeled.judge_ID) + 1

    # Choose random judges by id
    selected = npr.choice(np.arange(id_max), int(id_max/2), False)

    test_labeled = compas_labeled[compas_labeled.judge_ID.isin(selected)]
    train = compas_labeled[~compas_labeled.judge_ID.isin(selected)]

    # Renew IDs
    test_labeled.loc[:, 'judge_ID'] = test_labeled.groupby('judge_ID').ngroup()

    # Assign same observations to unlabeled data
    test_unlabeled = compas_unlabeled.iloc[test_labeled.index.values]

    # Train a logistic regression model
    B_model, predictions = fitPredictiveModel(
        train.loc[train['decision_T'] == 1, feature_cols],
        train.loc[train['decision_T'] == 1, 'result_Y'], test_labeled[feature_cols], 0)

    # Attach predictions to data
    test_labeled = test_labeled.assign(B_prob_0_model=predictions)
    test_unlabeled = test_unlabeled.assign(B_prob_0_model=predictions)

    test_labeled.sort_values(by='B_prob_0_model', inplace=True, ascending=True)

    # Find observed values
    observed = test_labeled['decision_T'] == 1

    # Assign data to the model
    dat = dict(D=int(10),
               N_obs=int(np.sum(observed)),
               N_cens=int(np.sum(~observed)),
               sigma_tau=sigma_tau,
               M=int(len(set(test_labeled['judge_ID']))),
               jj_obs=test_labeled.loc[observed, 'judge_ID']+1,
               jj_cens=test_labeled.loc[~observed, 'judge_ID']+1,
               dec_obs=test_labeled.loc[observed, 'decision_T'],
               dec_cens=test_labeled.loc[~observed, 'decision_T'],
               X_obs=test_labeled.loc[observed, feature_cols.values].values,
               X_cens=test_labeled.loc[~observed, feature_cols.values].values,
               y_obs=test_labeled.loc[observed, 'result_Y'].astype(int))


    # Do the sampling
    fit = sm.sample(data=dat, chains=4, seed=123, warmup_iters=1500,
                        sampling_iters=4000, adapt_delta=0.975)

    # Extract predictions
    fit_colnames = fit.column_names

    pred_cols = [s for s in fit_colnames if 'y_est' in s]

    pars = fit.get_drawset(params=pred_cols)

    fit.get_drawset(params=pred_cols)

    plt.close('all')
    gc.collect()
    gc.collect()

    print(fit.diagnose(), file=open(save_name + '_stan_fit_diagnostics_' +
          str(i) + '.txt', 'w'))

    # Draw figures for every 5th iteration
    if i % 5 == 0:
        # Save fit object to csv
        fit.save_csvfiles(dir=save_name, basename='fit_obj_i_' + str(i))
    # Bayes

    # Format matrix, each row is a sample of the posterior
    # columns are observations

    #y_imp = np.ones((pars['y_est'].shape[0], test_labeled.shape[0]))
    y_imp = np.ones((pars.shape[0], test_labeled.shape[0]))

    # Impute the unobserved with the estimated
    # Reverse the binary coding to compute failures as sum. 0 == 0 = 1
    #y_imp[:, ~observed] = 1 - pars['y_est']
    y_imp[:, ~observed] = 1 - pars

    pars = None
    fit = None

    del(pars)
    del(fit)

    gc.collect()
    gc.collect()

    process = psutil.Process(os.getpid())
    print("Memory use after deletion:", process.memory_info().rss)  # in bytes

    # Impute the observed
    y_imp[:, observed] = 1 - test_labeled.loc[observed, 'result_Y']


    for r in range(1, 9):

        to_release = np.round(test_labeled.shape[0] * (r / 10)).astype(int)

        failed = np.sum(y_imp[:, 0:to_release], axis=1)

        f_rate_cf[i, r - 1] = np.mean(failed / test_labeled.shape[0])

        print(".", end="")

        # True evaluation

        f_rate_true[i, r - 1], N = trueEvaluationEvaluator(test_unlabeled,
                   'result_Y', r / 10)

        # Labeled outcomes only

        f_rate_label[i, r - 1], N = labeledOutcomesEvaluator(test_labeled,
                    'decision_T', 'result_Y', r / 10)

        # Contraction

        f_rate_cont[i, r - 1], N = contraction(test_labeled, 'judge_ID',
                   'decision_T', 'result_Y', 'B_prob_0_model',
                   'judge_leniency', r / 10)

failure_rates[:, 0] = np.nanmean(f_rate_true, axis=0)
failure_rates[:, 1] = np.nanmean(f_rate_label, axis=0)
failure_rates[:, 2] = np.nanmean(f_rate_cont, axis=0)
failure_rates[:, 3] = np.nanmean(f_rate_cf, axis=0)

failure_stds[:, 0] = np.nanstd(f_rate_true, axis=0, ddof=1)
failure_stds[:, 1] = np.nanstd(f_rate_label, axis=0, ddof=1)
failure_stds[:, 2] = np.nanstd(f_rate_cont, axis=0, ddof=1)
failure_stds[:, 3] = np.nanstd(f_rate_cf, axis=0, ddof=1)

x_ax = np.arange(0.1, 0.9, 0.1)

labels = ['True Evaluation', 'Labeled outcomes', 'Contraction',
          'Counterfactuals']

colours = ['g', 'magenta', 'b', 'r']

line_styles = ['--', ':', '-.', '-']

for i in range(failure_rates.shape[1]):
    plt.errorbar(x_ax,
                 failure_rates[:, i],
                 label=labels[i],
                 c=colours[i],
                 linestyle=line_styles[i],
                 yerr=failure_stds[:, i])

#plt.title('FR vs. AR')
plt.xlabel('Acceptance rate')
plt.ylabel('Failure rate')
plt.legend()
plt.grid()

plt.savefig(save_name + '_all')

plt.show()

plt.close('all')

# Plot error of the methods (contraction and counterfactual)
# w.r.t True evaluation
for i in range(2, failure_rates.shape[1]):

    error = np.abs(failure_rates[:, 0] - failure_rates[:, i])

    plt.errorbar(x_ax,
                 error,
                 label=labels[i],
                 c=colours[i],
                 linestyle=line_styles[i],
                 yerr=failure_stds[:, i])

plt.axhline(y=0, c='k')
plt.xlabel('Acceptance rate')
plt.ylabel('Error w.r.t. True evaluation')
plt.legend()
plt.grid()

plt.savefig(save_name + '_all_err')

plt.show()

plt.close('all')

print("\nSingle runs:")
print(f_rate_true, f_rate_label, f_rate_cont, f_rate_cf)

print("\nFailure rates:")
print(np.array2string(failure_rates, formatter={'float_kind':lambda x: "%.5f" % x}))

print("\nStandard deviations:")
print(np.array2string(failure_stds, formatter={'float_kind':lambda x: "%.5f" % x}))

# Save arrays
np.save(save_name + "_true_FRs", f_rate_true)
np.save(save_name + "_labeled_FRs", f_rate_label)
np.save(save_name + "_contraction_FRs", f_rate_cont)
np.save(save_name + "_counterfactuals_FRs", f_rate_cf)

print("\nMean absolute errors:")
for i in range(1, failure_rates.shape[1]):
    print(
        labels[i].ljust(len(max(labels, key=len))),
        np.round(
            np.mean(np.abs(failure_rates[:, 0] - failure_rates[:, i])), 5))

drawDiagnostics(title="Variation of failure rate estimates by method\nCOMPAS data set",
                save_name=save_name,
                f_rates=[f_rate_true, f_rate_label, f_rate_cont, f_rate_cf],
                titles=labels)
